from snowddl.blueprint import ExternalFunctionBlueprint, Ident, SchemaObjectIdentWithArgs, NameWithType, DataType, build_schema_object_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


external_function_json_schema = {
    "type": "object",
    "properties": {
        "arguments": {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            }
        },
        "returns": {
            "type": "string"
        },
        "api_integration": {
            "type": "string"
        },
        "url": {
            "type": "string",
        },
        "is_secure": {
            "type": "boolean"
        },
        "is_strict": {
            "type": "boolean"
        },
        "is_immutable": {
            "type": "boolean"
        },
        "headers": {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            }
        },
        "context_headers": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "max_batch_rows": {
            "type": "integer"
        },
        "compression": {
            "type": "string"
        },
        "request_translator": {
            "type": "string"
        },
        "response_translator": {
            "type": "string"
        },
        "comment": {
            "type": "string"
        }
    },
    "required": ["api_integration", "url"],
    "additionalProperties": False
}


class ExternalFunctionParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("external_function", external_function_json_schema, self.process_function)

    def process_function(self, f: ParsedFile):
        arguments = [NameWithType(name=Ident(k), type=DataType(t)) for k, t in f.params.get('arguments', {}).items()]
        returns = DataType(f.params['returns'])

        base_name = self.validate_name_with_args(f.path, arguments)

        bp = ExternalFunctionBlueprint(
            full_name=SchemaObjectIdentWithArgs(self.env_prefix, f.database, f.schema, base_name, [a.type.base_type for a in arguments]),
            arguments=arguments,
            returns=returns,
            api_integration=Ident(f.params.get('api_integration')),
            url=f.params['url'],
            is_secure=f.params.get('is_secure', False),
            is_strict=f.params.get('is_strict', False),
            is_immutable=f.params.get('is_immutable', False),
            headers=f.params.get('headers'),
            context_headers=[Ident(h) for h in f.params.get('context_headers')] if f.params.get('context_headers') else None,
            max_batch_rows=f.params.get('max_batch_rows'),
            compression=f.params.get('compression'),
            request_translator=build_schema_object_ident(self.env_prefix, f.params['request_translator'], f.database, f.schema) if f.params.get('request_translator') else None,
            response_translator=build_schema_object_ident(self.env_prefix, f.params['response_translator'], f.database, f.schema) if f.params.get('response_translator') else None,
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import DynamicTableBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class DynamicTableResolver(AbstractSchemaObjectResolver):
    # Dynamic tables are available for all accounts during preview, including STANDARD edition
    # skip_min_edition = Edition.ENTERPRISE
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.DYNAMIC_TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW DYNAMIC TABLES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                # Extract SQL query text only, skip the initial "CREATE DYNAMIC TABLE ..." part
                # Snowflake modifies original SQL text in this column, it cannot be compared directly
                "text": r['text'].partition("\nAS\n")[2],
                "target_lag": r['target_lag'],
                "refresh_mode": r['refresh_mode'],
                "warehouse": r['warehouse'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(DynamicTableBlueprint)

    def create_object(self, bp: DynamicTableBlueprint):
        query = self.engine.query_builder()
        query.append("CREATE DYNAMIC TABLE {full_name:i}", {
            "full_name": bp.full_name,
        })

        query.append(self._build_common_dynamic_table_sql(bp))

        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: DynamicTableBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if bp.text != row['text']:
            query = self.engine.query_builder()
            query.append("CREATE OR REPLACE DYNAMIC TABLE {full_name:i}", {
                "full_name": bp.full_name,
            })

            query.append(self._build_common_dynamic_table_sql(bp))
            self.engine.execute_unsafe_ddl(query)

            return ResolveResult.REPLACE

        if bp.target_lag != row['target_lag']:
            self.engine.execute_safe_ddl("ALTER DYNAMIC TABLE {full_name:i} SET TARGET_LAG = {target_lag}", {
                "full_name": bp.full_name,
                "target_lag": bp.target_lag,
            })

            result = ResolveResult.ALTER

        if str(bp.warehouse) != row['warehouse']:
            self.engine.execute_safe_ddl("ALTER DYNAMIC TABLE {full_name:i} SET WAREHOUSE = {warehouse:i}", {
                "full_name": bp.full_name,
                "warehouse": bp.warehouse,
            })

            result = ResolveResult.ALTER

        if bp.comment != row['comment']:
            self.engine.execute_safe_ddl("ALTER DYNAMIC TABLE {full_name:i} SET COMMENT = {comment}", {
                "full_name": bp.full_name,
                "comment": bp.comment,
            })

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP DYNAMIC TABLE {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        return ResolveResult.DROP

    def _build_common_dynamic_table_sql(self, bp: DynamicTableBlueprint):
        query = self.engine.query_builder()

        query.append_nl("TARGET_LAG = {target_lag}", {
            "target_lag": bp.target_lag,
        })

        query.append_nl("WAREHOUSE = {warehouse:i}", {
            "warehouse": bp.warehouse,
        })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        query.append_nl("AS")
        query.append_nl(bp.text)

        return query



from snowddl.blueprint import StageBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import coalesce, compare_dynamic_param_value

class StageResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.STAGE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW STAGES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            if 'TEMPORARY' in r['type']:
                # Skip TEMPORARY stages
                continue

            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "url": r['url'] if r['url'] else None,
                "storage_integration": r['storage_integration'] if r['storage_integration'] else None,
                "owner": r['owner'],
                "type": r['type'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(StageBlueprint)

    def create_object(self, bp: StageBlueprint):
        query = self.engine.query_builder()
        query.append("CREATE STAGE {full_name:i}", {
            "full_name": bp.full_name,
        })

        query.append_nl(self._build_create_stage_sql(bp))
        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: StageBlueprint, row: dict):
        if self._is_replace_required(bp, row):
            query = self.engine.query_builder()
            query.append("CREATE OR REPLACE STAGE {full_name:i}", {
                "full_name": bp.full_name,
            })

            query.append_nl(self._build_create_stage_sql(bp))
            self.engine.execute_unsafe_ddl(query)

            return ResolveResult.REPLACE

        # Check specific properties for potential ALTER
        existing_properties = self._get_existing_stage_properties(bp)
        result = ResolveResult.NOCHANGE

        if self._compare_url(bp, row):
            result = ResolveResult.ALTER

        if self._compare_storage_integration(bp, row):
            result = ResolveResult.ALTER

        if self._compare_directory(bp, existing_properties):
            result = ResolveResult.ALTER

        if self._compare_file_format(bp, existing_properties):
            result = ResolveResult.ALTER

        if self._compare_copy_options(bp, existing_properties):
            result = ResolveResult.ALTER

        if self._compare_comment(bp, row):
            result = ResolveResult.ALTER

        if self.engine.settings.refresh_stage_encryption and bp.url and self._refresh_encryption(bp):
            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP STAGE {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        return ResolveResult.DROP

    def _is_replace_required(self, bp: StageBlueprint, row):
        # Change of stage type from INTERNAL to EXTERNAL or vice versa
        if (bp.url and not row['url']) or (not bp.url and row['url']):
            return True

        if bp.url is None:
            encryption_type = coalesce(bp.encryption, {}).get("TYPE", "SNOWFLAKE_FULL")

            if row['type'] == "INTERNAL" and encryption_type != "SNOWFLAKE_FULL":
                return True

            if row['type'] == "INTERNAL NO CSE" and encryption_type != "SNOWFLAKE_SSE":
                return True

        return False

    def _compare_url(self, bp: StageBlueprint, row):
        if not ((bp.url is None and row['url'] is None) or str(bp.url) == row['url']):
            query = self.engine.query_builder()
            query.append("ALTER STAGE {full_name:i} SET", {
                "full_name": bp.full_name,
            })

            query.append_nl("URL = {url:i}", {
                "url": bp.url,
            })

            self.engine.execute_safe_ddl(query)

            return True

        return False

    def _compare_storage_integration(self, bp: StageBlueprint, row):
        if not ((bp.storage_integration is None and row['storage_integration'] is None) or str(bp.storage_integration) == row['storage_integration']):
            query = self.engine.query_builder()
            query.append("ALTER STAGE {full_name:i} SET", {
                "full_name": bp.full_name,
            })

            query.append_nl("STORAGE_INTEGRATION = {storage_integration:i}", {
                "storage_integration": bp.storage_integration,
            })

            self.engine.execute_safe_ddl(query)

            return True

        return False

    def _compare_directory(self, bp: StageBlueprint, existing_properties):
        is_alter_required = False

        # Check properties in blueprint against DESC output
        for prop_name, prop_value in coalesce(bp.directory, {}).items():
            # REFRESH_ON_CREATE is only used for CREATE command, but is not available for ALTER, must be skipped
            if prop_name == "REFRESH_ON_CREATE":
                continue

            if prop_name not in existing_properties['DIRECTORY']:
                raise ValueError(f"Unknown DIRECTORY property [{prop_name}] for stage [{bp.full_name}]")

            if compare_dynamic_param_value(prop_value,  existing_properties['DIRECTORY'][prop_name]['property_value']):
                continue

            is_alter_required = True
            break

        # Check non-default properties in DESC output exist in blueprint
        for prop_name, prop in existing_properties['DIRECTORY'].items():
            # LAST_REFRESHED_ON is an informational property, it should not be compared or set explicitly
            if prop_name == "LAST_REFRESHED_ON":
                continue

            if prop['property_value'] != prop['property_default'] and (prop_name not in coalesce(bp.directory, {})):
                is_alter_required = True
                break

        # If anything did not match, refresh all properties for this section
        if is_alter_required:
            query = self.engine.query_builder()
            query.append("ALTER STAGE {full_name:i} SET", {
                "full_name": bp.full_name,
            })

            query.append_nl(self._build_directory(bp))
            self.engine.execute_safe_ddl(query)

        return is_alter_required

    def _compare_file_format(self, bp: StageBlueprint, existing_properties):
        # Only named FILE FORMATs are supported
        if 'FORMAT_NAME' in existing_properties['STAGE_FILE_FORMAT']:
            existing_file_format = str(existing_properties['STAGE_FILE_FORMAT']['FORMAT_NAME']['property_value'])
            existing_file_format = existing_file_format.replace('\\', '').replace('"', '')
        else:
            existing_file_format = None

        if not ((bp.file_format is None and existing_file_format is None) or str(bp.file_format) == existing_file_format):
            query = self.engine.query_builder()
            query.append("ALTER STAGE {full_name:i} SET", {
                "full_name": bp.full_name,
            })

            query.append_nl(self._build_file_format(bp))
            self.engine.execute_safe_ddl(query)

            return True

        return False

    def _compare_copy_options(self, bp: StageBlueprint, existing_properties):
        is_alter_required = False

        # Check properties in blueprint against DESC output
        for prop_name, prop_value in coalesce(bp.copy_options, {}).items():
            if prop_name not in existing_properties['STAGE_COPY_OPTIONS']:
                raise ValueError(f"Unknown COPY_OPTIONS property [{prop_name}] for stage [{bp.full_name}]")

            if compare_dynamic_param_value(prop_value,  existing_properties['STAGE_COPY_OPTIONS'][prop_name]['property_value']):
                continue

            is_alter_required = True
            break

        # Check non-default properties in DESC output exist in blueprint
        for prop_name, prop in existing_properties['STAGE_COPY_OPTIONS'].items():
            if prop['property_value'] != prop['property_default'] and (prop_name not in coalesce(bp.copy_options, {})):
                # ENFORCE_LENGTH is a mirror for TRUNCATECOLUMNS
                if prop_name == "ENFORCE_LENGTH" and "TRUNCATECOLUMNS" in coalesce(bp.copy_options, {}):
                    continue

                # TRUNCATECOLUMNS is a mirror for ENFORCE_LENGTH
                if prop_name == "TRUNCATECOLUMNS" and "ENFORCE_LENGTH" in coalesce(bp.copy_options, {}):
                    continue

                is_alter_required = True
                break

        # If anything did not match, refresh all properties for this section
        if is_alter_required:
            query = self.engine.query_builder()
            query.append("ALTER STAGE {full_name:i} SET", {
                "full_name": bp.full_name,
            })

            query.append_nl(self._build_copy_options(bp))
            self.engine.execute_safe_ddl(query)

        return is_alter_required

    def _compare_comment(self, bp: StageBlueprint, row):
        if not ((bp.comment is None and row['comment'] is None) or str(bp.comment) == row['comment']):
            query = self.engine.query_builder()
            query.append("ALTER STAGE {full_name:i} SET", {
                "full_name": bp.full_name,
            })

            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

            self.engine.execute_safe_ddl(query)

            return True

        return False

    def _refresh_encryption(self, bp: StageBlueprint):
        if bp.encryption:
            query = self.engine.query_builder()
            query.append("ALTER STAGE {full_name:i} SET", {
                "full_name": bp.full_name,
            })

            query.append_nl(self._build_encryption(bp))

            self.engine.execute_safe_ddl(query)

            return True

        return False

    def _build_directory(self, bp: StageBlueprint):
        query = self.engine.query_builder()

        if bp.directory:
            query.append("DIRECTORY = (")

            for k, v in bp.directory.items():
                query.append("{param_name:r} = {param_value:dp}", {
                    "param_name": k,
                    "param_value": v,
                })

            query.append(")")
        else:
            query.append("DIRECTORY = ( ENABLE = FALSE )")

        return query

    def _build_file_format(self, bp: StageBlueprint):
        query = self.engine.query_builder()

        if bp.file_format:
            query.append("FILE_FORMAT = ( FORMAT_NAME = {file_format:i} )", {
                "file_format": bp.file_format,
            })
        else:
            query.append("FILE_FORMAT = ( TYPE = CSV )")

        return query

    def _build_copy_options(self, bp: StageBlueprint):
        query = self.engine.query_builder()

        if bp.copy_options:
            query.append("COPY_OPTIONS = (")

            for k, v in bp.copy_options.items():
                query.append("{param_name:r} = {param_value:dp}", {
                    "param_name": k,
                    "param_value": v,
                })

            query.append(")")
        else:
            query.append("COPY_OPTIONS = ()")

        return query

    def _build_encryption(self, bp: StageBlueprint):
        query = self.engine.query_builder()

        if bp.encryption:
            query.append("ENCRYPTION = (")

            for k, v in bp.encryption.items():
                query.append("{param_name:r} = {param_value:dp}", {
                    "param_name": k,
                    "param_value": v,
                })

            query.append(")")

        # Cannot reset ENCRYPTION properties to default with ENCRYPTION = () or UNSET
        # It must be defined explicitly in config for any changes to take effect

        return query

    def _build_create_stage_sql(self, bp: StageBlueprint):
        query = self.engine.query_builder()

        if bp.url:
            query.append_nl("URL = {url}", {
                "url": bp.url
            })

        if bp.storage_integration:
            query.append_nl("STORAGE_INTEGRATION = {storage_integration:i}", {
                "storage_integration": bp.storage_integration,
            })

        if bp.directory:
            query.append_nl(self._build_directory(bp))

        if bp.file_format:
            query.append_nl(self._build_file_format(bp))

        if bp.copy_options:
            query.append_nl(self._build_copy_options(bp))

        if bp.encryption:
            query.append_nl(self._build_encryption(bp))

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        return query

    def _get_existing_stage_properties(self, bp: StageBlueprint):
        cur = self.engine.execute_meta("DESC STAGE {full_name:i}", {
            "full_name": bp.full_name,
        })

        result = {
            "STAGE_FILE_FORMAT": {},
            "STAGE_COPY_OPTIONS": {},
            "DIRECTORY": {},
        }

        for r in cur:
            if r['parent_property'] not in result:
                result[r['parent_property']] = {}

            result[r['parent_property']][r['property']] = r

        return result



from abc import abstractmethod

from snowddl.blueprint import RoleBlueprint, Grant, FutureGrant, SchemaIdent, SchemaObjectIdent, build_grant_name_ident_snowflake
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class AbstractRoleResolver(AbstractResolver):
    @abstractmethod
    def get_role_suffix(self) -> str:
        pass

    def get_object_type(self):
        return ObjectType.ROLE

    def get_existing_objects(self):
        existing_roles = {}

        cur = self.engine.execute_meta("SHOW ROLES LIKE {pattern:lse}", {
            'pattern': (self.config.env_prefix, f"__{self.get_role_suffix()}"),
        })

        for r in cur:
            if r['owner'] != self.engine.context.current_role:
                continue

            existing_roles[r['name']] = {
                "role_name": r['name'],
                "comment": r['comment'] if r['comment'] else None,
            }

        # Process role grants in parallel
        for role_name, grants, future_grants in self.engine.executor.map(self.get_existing_role_grants, existing_roles):
            existing_roles[role_name]['grants'] = grants
            existing_roles[role_name]['future_grants'] = future_grants

        return existing_roles

    def get_existing_role_grants(self, role_name):
        grants = []
        future_grants = []

        cur = self.engine.execute_meta("SHOW GRANTS TO ROLE {role_name:i}", {
            "role_name": role_name,
        })

        for r in sorted(cur, key=self.sort_existing_grants):
            # Skip grants on unknown object types
            try:
                object_type = ObjectType[r['granted_on']]
            except KeyError:
                continue

            # Snowflake bug: phantom MATERIALIZED VIEW when SEARCH OPTIMIZATION is enabled for table
            if object_type == ObjectType.MATERIALIZED_VIEW and str(r['name']).endswith('IDX_MV_"'):
                continue

            grants.append(Grant(
                privilege=r['privilege'],
                on=object_type,
                name=build_grant_name_ident_snowflake(r['name'], object_type),
            ))

        cur = self.engine.execute_meta("SHOW FUTURE GRANTS TO ROLE {role_name:i}", {
            "role_name": role_name,
        })

        for r in sorted(cur, key=self.sort_existing_grants):
            # Skip future grants on unknown object types
            try:
                object_type = ObjectType[r['grant_on']]
            except KeyError:
                continue

            future_grants.append(FutureGrant(
                privilege=r['privilege'],
                on=object_type,
                name=build_grant_name_ident_snowflake(r['name'], object_type),
            ))

        return role_name, grants, future_grants

    def sort_existing_grants(self, row):
        offset = 0

        # WRITE should be revoked before READ
        if row['privilege'] == "WRITE":
            offset = -1

        return (row.get('grant_on'), row.get('granted_on'), row['name'], offset, row['privilege'])

    def create_object(self, bp: RoleBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE ROLE {role_name:i}", {
            "role_name": bp.full_name,
        })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl("GRANT ROLE {role_name:i} TO ROLE {current_role:i}", {
            "role_name": bp.full_name,
            "current_role": self.engine.context.current_role,
        })

        for bp_grant in bp.grants:
            self.create_grant(bp.full_name, bp_grant)

        for bp_future_grant in bp.future_grants:
            self.create_future_grant(bp.full_name, bp_future_grant)

        return ResolveResult.CREATE

    def compare_object(self, bp: RoleBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if bp.comment != row['comment']:
            self.engine.execute_safe_ddl("ALTER ROLE {role_name:i} SET COMMENT = {comment}", {
                "role_name": bp.full_name,
                "comment": bp.comment,
            })

            result = ResolveResult.ALTER

        for bp_grant in bp.grants:
            if bp_grant not in row['grants']:
                self.create_grant(bp.full_name, bp_grant)
                result = ResolveResult.GRANT

        for existing_grant in row['grants']:
            if existing_grant not in bp.grants \
            and self.grant_to_future_grant(existing_grant) not in bp.future_grants:
                self.drop_grant(bp.full_name, existing_grant)
                result = ResolveResult.GRANT

        for bp_future_grant in bp.future_grants:
            if bp_future_grant not in row['future_grants']:
                self.create_future_grant(bp.full_name, bp_future_grant)
                result = ResolveResult.GRANT

            if self.engine.settings.refresh_future_grants:
                self.refresh_future_grant(bp.full_name, bp_future_grant)
                result = ResolveResult.GRANT

        for existing_future_grant in row['future_grants']:
            if existing_future_grant not in bp.future_grants:
                self.drop_future_grant(bp.full_name, existing_future_grant)
                result = ResolveResult.GRANT

        return result

    def drop_object(self, row: dict):
        # All grants and future grants are destroyed automatically
        self.engine.execute_unsafe_ddl("DROP ROLE {role_name:i}", {
            "role_name": row['role_name'],
        })

        return ResolveResult.DROP

    def create_grant(self, role_name, grant: Grant):
        if grant.privilege == "USAGE" and grant.on == ObjectType.ROLE:
            self.engine.execute_safe_ddl("GRANT ROLE {name:i} TO ROLE {role_name:i}", {
                "name": grant.name,
                "role_name": role_name,
            })
        else:
            self.engine.execute_safe_ddl("GRANT {privilege:r} ON {on:r} {name:i} TO ROLE {role_name:i}", {
                "privilege": grant.privilege,
                "on": grant.on.singular,
                "name": grant.name,
                "role_name": role_name,
            })

    def drop_grant(self, role_name, grant: Grant):
        # OWNERSHIP can only be transferred
        if grant.privilege == "OWNERSHIP":
            return

        if grant.privilege == "USAGE" and grant.on == ObjectType.ROLE:
            self.engine.execute_safe_ddl("REVOKE ROLE {name:i} FROM ROLE {role_name:i}", {
                "name": grant.name,
                "role_name": role_name,
            })
        else:
            self.engine.execute_safe_ddl("REVOKE {privilege:r} ON {on:r} {name:i} FROM ROLE {role_name:i}", {
                "privilege": grant.privilege,
                "on": grant.on.singular,
                "name": grant.name,
                "role_name": role_name,
            })

    def create_future_grant(self, role_name, grant: FutureGrant):
        self.engine.execute_safe_ddl("GRANT {privilege:r} ON FUTURE {on_plural:r} IN SCHEMA {name:i} TO ROLE {role_name:i}", {
            "privilege": grant.privilege,
            "on_plural": grant.on.plural,
            "name": grant.name,
            "role_name": role_name,
        })

    def drop_future_grant(self, role_name, grant: FutureGrant):
        self.engine.execute_safe_ddl("REVOKE {privilege:r} ON FUTURE {on_plural:r} IN SCHEMA {name:i} FROM ROLE {role_name:i}", {
            "privilege": grant.privilege,
            "on_plural": grant.on.plural,
            "name": grant.name,
            "role_name": role_name,
        })

    def refresh_future_grant(self, role_name, grant: FutureGrant):
        # Bulk grant on objects of type PIPE to ROLE is restricted (by Snowflake)
        if grant.on == ObjectType.PIPE:
            return

        self.engine.execute_safe_ddl("GRANT {privilege:r} ON ALL {on_plural:r} IN SCHEMA {name:i} TO ROLE {role_name:i}", {
            "privilege": grant.privilege,
            "on_plural": grant.on.plural,
            "name": grant.name,
            "role_name": role_name,
        })

    def grant_to_future_grant(self, grant: Grant):
        if not grant.on.is_future_grant_supported:
            return None

        future_grant = FutureGrant(
            privilege=grant.privilege,
            on=grant.on,
            name=SchemaIdent(grant.name.env_prefix, grant.name.database, grant.name.schema) if isinstance(grant.name, SchemaObjectIdent) else grant.name
        )

        return future_grant



from snowddl.blueprint import RoleBlueprint, UserBlueprint, Grant, build_role_ident, build_grant_name_ident_snowflake
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class UserRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.USER_ROLE_SUFFIX

    def get_existing_role_grants(self, role_name):
        grants = []

        cur = self.engine.execute_meta("SHOW GRANTS TO ROLE {role_name:i}", {
            "role_name": role_name,
        })

        for r in cur:
            # Check ROLE grants only, ignore everything else
            # User roles may accumulate random grants from temporary tables and stages
            # as well as ownership of manually created objects
            if r['granted_on'] != 'ROLE':
                continue

            grants.append(Grant(
                privilege=r['privilege'],
                on=ObjectType[r['granted_on']],
                name=build_grant_name_ident_snowflake(r['name'], ObjectType[r['granted_on']]),
            ))

        return role_name, grants, []

    def get_blueprints(self):
        blueprints = []

        for user in self.config.get_blueprints_by_type(UserBlueprint).values():
            blueprints.append(self.get_blueprint_user_role(user))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_user_role(self, user: UserBlueprint):
        grants = []

        for business_role in user.business_roles:
            grants.append(Grant(
                privilege="USAGE",
                on=ObjectType.ROLE,
                name=business_role,
            ))

        bp = RoleBlueprint(
            full_name=build_role_ident(self.config.env_prefix, user.full_name, self.get_role_suffix()),
            grants=grants,
            future_grants=[],
            comment=None,
        )

        return bp



from snowddl.blueprint import AlertBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class AlertResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.ALERT

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW ALERTS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "warehouse": r['warehouse'],
                "schedule": r['schedule'],
                "state": r['state'],
                "condition": r['condition'],
                "action": r['action'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(AlertBlueprint)

    def create_object(self, bp: AlertBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE ALERT {full_name:i}", {
            "full_name": bp.full_name,
        })

        query.append_nl("WAREHOUSE = {warehouse:i}", {
            "warehouse": bp.warehouse,
        })

        query.append_nl("SCHEDULE = {schedule}", {
            "schedule": bp.schedule,
        })

        query.append_nl("IF(EXISTS(")
        query.append_nl(bp.condition)
        query.append_nl("))")

        query.append_nl("THEN")
        query.append_nl(bp.action)

        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: AlertBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if str(bp.warehouse) != row['warehouse']:
            self.engine.execute_safe_ddl("ALTER ALERT {full_name:i} SET warehouse = {warehouse:i}", {
                "full_name": bp.full_name,
                "warehouse": bp.warehouse,
            })

            result = ResolveResult.ALTER

        if str(bp.schedule) != row['schedule']:
            self.engine.execute_safe_ddl("ALTER ALERT {full_name:i} SET schedule = {schedule}", {
                "full_name": bp.full_name,
                "schedule": bp.schedule,
            })

            result = ResolveResult.ALTER

        if str(bp.condition) != row['condition']:
            query = self.engine.query_builder()

            query.append("ALTER ALERT {full_name:i} MODIFY CONDITION EXISTS (", {
                "full_name": bp.full_name,
            })

            query.append_nl(bp.condition)
            query.append_nl(")")

            self.engine.execute_safe_ddl(query)

            result = ResolveResult.ALTER

        if str(bp.action) != row['action']:
            query = self.engine.query_builder()

            query.append("ALTER ALERT {full_name:i} MODIFY ACTION", {
                "full_name": bp.full_name,
            })

            query.append_nl(bp.action)

            self.engine.execute_safe_ddl(query)

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl("DROP ALERT {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        return ResolveResult.DROP



from snowddl.blueprint import TableBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractResolver, ResolveResult, ObjectType


class CloneTableResolver(AbstractResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.CLONE_TABLE

    def get_blueprints(self):
        return {}

    def get_existing_objects(self):
        databases_for_clone = self.get_databases_for_clone()
        schemas_for_clone = {}
        tables_for_clone = {}

        # Get schemas for clone in parallel
        for schema_objects in self.engine.executor.map(self.get_schemas_for_clone, databases_for_clone.values()):
            schemas_for_clone.update(schema_objects)

        # Get tables for clone in parallel
        for table_objects in self.engine.executor.map(self.get_tables_for_clone, schemas_for_clone.values()):
            tables_for_clone.update(table_objects)

        return tables_for_clone

    def get_databases_for_clone(self):
        databases_for_clone = {}

        cur = self.engine.execute_meta("SHOW DATABASES")

        for r in cur:
            # Skip shares
            if r['origin']:
                continue

            # Skip databases without destination for cloning
            if f"{self.config.env_prefix}{r['name']}" not in self.engine.schema_cache.databases:
                continue

            databases_for_clone[r['name']] = {
                "database": r['name'],
            }

        return databases_for_clone

    def get_schemas_for_clone(self, database):
        schemas_for_clone = {}

        cur = self.engine.execute_meta("SHOW SCHEMAS IN DATABASE {database:i}", {
            "database": database['database'],
        })

        for r in cur:
            # Skip INFORMATION_SCHEMA
            if r['name'] == "INFORMATION_SCHEMA":
                continue

            # Skip schemas without destination for cloning
            if f"{self.config.env_prefix}{r['database_name']}.{r['name']}" not in self.engine.schema_cache.schemas:
                continue

            schemas_for_clone[f"{r['database_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['name'],
            }

        return schemas_for_clone

    def get_tables_for_clone(self, schema):
        tables_for_clone = {}

        cur = self.engine.execute_meta("SHOW TABLES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            # Skip external tables
            if r['is_external'] == 'Y':
                continue

            tables_for_clone[f"{self.config.env_prefix}{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "is_transient": r['kind'] == 'TRANSIENT',
            }

        return tables_for_clone

    def create_object(self, bp: TableBlueprint):
        return ResolveResult.NOCHANGE

    def compare_object(self, bp: TableBlueprint, row: dict):
        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        # Drop object is utilized since we have no blueprints for clones, but only existing objects
        # All such objects should be "dropped" from AbstractResolver perspective
        # TODO: find more elegant solution if cloning becomes popular

        query = self.engine.query_builder()
        query.append("CREATE")

        if row['is_transient']:
            query.append("TRANSIENT")

        query.append("TABLE IF NOT EXISTS {database_with_prefix:i}.{schema:i}.{table_name:i}", {
            "database_with_prefix": f"{self.config.env_prefix}{row['database']}",
            "schema": row['schema'],
            "table_name": row['name'],
        })

        query.append_nl("CLONE {database:i}.{schema:i}.{table_name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "table_name": row['name'],
        })

        cur = self.engine.execute_clone(query)
        r = cur.fetchone()

        if str(r['status']).endswith("successfully created."):
            return ResolveResult.CREATE

        return ResolveResult.NOCHANGE

    def destroy(self):
        pass

    def _is_skipped(self):
        if not self.engine.settings.clone_table:
            return True

        if super()._is_skipped():
            return True

        return False



from snowddl.blueprint import EventTableBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class EventTableResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.EVENT_TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW EVENT TABLES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"
            existing_objects[full_name] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "retention_time": int(r['retention_time']),
                "cluster_by": r['cluster_by'] if r['cluster_by'] else None,
                "change_tracking": bool(r['change_tracking'] == 'ON'),
                "search_optimization": bool(r.get('search_optimization') == 'ON'),
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(EventTableBlueprint)

    def create_object(self, bp: EventTableBlueprint):
        query = self._build_create_table(bp)
        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: EventTableBlueprint, row: dict):
        safe_alters = []
        unsafe_alters = []

        # Change tracking
        if bp.change_tracking != row['change_tracking']:
            unsafe_alters.append(self.engine.format("SET CHANGE_TRACKING = {change_tracking:b}", {
                "change_tracking": bp.change_tracking,
            }))

        # Comment
        if bp.comment != row['comment']:
            if bp.comment:
                safe_alters.append(self.engine.format("SET COMMENT = {comment}", {
                    "comment": bp.comment,
                }))
            else:
                safe_alters.append(self.engine.format("UNSET COMMENT"))

        # Apply changes
        result = ResolveResult.NOCHANGE

        if safe_alters or unsafe_alters:
            for alter in safe_alters:
                self.engine.execute_safe_ddl("ALTER TABLE {full_name:i} {alter:r}", {
                    "full_name": bp.full_name,
                    "alter": alter,
                })

            for alter in unsafe_alters:
                self.engine.execute_unsafe_ddl("ALTER TABLE {full_name:i} {alter:r}", {
                    "full_name": bp.full_name,
                    "alter": alter,
                })

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP TABLE {database:i}.{schema:i}.{table_name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "table_name": row['name'],
        })

        return ResolveResult.DROP

    def _build_create_table(self, bp: EventTableBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE EVENT TABLE {full_name:i}", {
            "full_name": bp.full_name,
        })

        if bp.change_tracking:
            query.append_nl("CHANGE_TRACKING = TRUE")

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        return query



from snowddl.blueprint import FunctionBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import dtypes_from_arguments


class FunctionResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.FUNCTION

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW USER FUNCTIONS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            # Skip external functions
            if r['is_external_function'] == 'Y':
                continue

            full_name = f"{r['catalog_name']}.{r['schema_name']}.{r['name']}({dtypes_from_arguments(r['arguments'])})"

            existing_objects[full_name] = {
                "database": r['catalog_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "arguments": r['arguments'],
                "comment": r['description'],
                "is_table_function": r['is_table_function'] == 'Y',
                "is_secure": r['is_secure'] == 'Y',
                "is_memoizable": r['is_memoizable'] == 'Y',
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(FunctionBlueprint)

    def create_object(self, bp: FunctionBlueprint):
        query = self._build_create_function(bp)

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl("COMMENT ON FUNCTION {full_name:i} IS {comment}", {
            "full_name": bp.full_name,
            "comment": query.add_short_hash(bp.comment),
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: FunctionBlueprint, row: dict):
        query = self._build_create_function(bp)

        if not query.compare_short_hash(row['comment']):
            self.engine.execute_safe_ddl(query)

            self.engine.execute_safe_ddl("COMMENT ON FUNCTION {full_name:i} IS {comment}", {
                "full_name": bp.full_name,
                "comment": query.add_short_hash(bp.comment),
            })

            return ResolveResult.REPLACE

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl("DROP FUNCTION {database:i}.{schema:i}.{name:i}({dtypes:r})", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
            "dtypes": dtypes_from_arguments(row['arguments']),
        })

        return ResolveResult.DROP

    def _build_create_function(self, bp: FunctionBlueprint):
        query = self.engine.query_builder()
        query.append("CREATE OR REPLACE")

        if bp.is_secure:
            query.append("SECURE")

        query.append("FUNCTION {full_name:in} (", {
            "full_name": bp.full_name,
        })

        for idx, arg in enumerate(bp.arguments):
            query.append_nl("    {comma:r}{arg_name:i} {arg_type:r}", {
                "comma": "  " if idx == 0 else ", ",
                "arg_name": arg.name,
                "arg_type": arg.type,
            })

        query.append_nl(")")

        if isinstance(bp.returns, list):
            query.append_nl("RETURNS TABLE (")

            for idx, arg in enumerate(bp.returns):
                query.append_nl("    {comma:r}{ret_name:i} {ret_type:r}", {
                    "comma": "  " if idx == 0 else ", ",
                    "ret_name": arg.name,
                    "ret_type": arg.type,
                })

            query.append_nl(")")
        else:
            query.append_nl("RETURNS {ret_type:r}", {
                "ret_type": bp.returns,
            })

        query.append_nl("LANGUAGE {language:r}", {
            "language": bp.language,
        })

        if bp.is_strict:
            query.append_nl("STRICT")

        if bp.is_immutable:
            query.append_nl("IMMUTABLE")

        if bp.is_memoizable:
            query.append_nl("MEMOIZABLE")

        if bp.runtime_version:
            query.append_nl("RUNTIME_VERSION = {runtime_version}", {
                "runtime_version": bp.runtime_version,
            })

        if bp.imports:
            query.append_nl("IMPORTS = ({imports})", {
                "imports": [f"@{i.stage_name}{i.path}" for i in bp.imports]
            })

        if bp.packages:
            query.append_nl("PACKAGES = ({packages})", {
                "packages": bp.packages,
            })

        if bp.handler:
            query.append_nl("HANDLER = {handler}", {
                "handler": bp.handler,
            })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        if bp.body:
            query.append_nl("AS {body}", {
                "body": bp.body,
            })

        return query



from snowddl.blueprint import AccountObjectIdent, DynamicTableBlueprint, SchemaObjectIdent, build_schema_object_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


dynamic_table_json_schema = {
    "type": "object",
    "properties": {
        "text": {
            "type": "string"
        },
        "target_lag": {
            "type": "string"
        },
        "warehouse": {
            "type": "string",
        },
        "depends_on": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "comment": {
            "type": "string"
        },
    },
    "required": ["text", "target_lag", "warehouse"],
    "additionalProperties": False
}


class DynamicTableParser(AbstractParser):
    unit_singular_to_plural = {
        "second": "seconds",
        "minute": "minutes",
        "hour": "hours",
        "day": "days",
    }

    unit_plural_to_singular = {v: k for k,v in unit_singular_to_plural.items()}

    def load_blueprints(self):
        self.parse_schema_object_files("dynamic_table", dynamic_table_json_schema, self.process_dynamic_table)

    def process_dynamic_table(self, f: ParsedFile):

        bp = DynamicTableBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            text=f.params['text'],
            target_lag=self.normalize_target_lag(f.params['target_lag']),
            warehouse=AccountObjectIdent(self.env_prefix, f.params['warehouse']),
            depends_on=[build_schema_object_ident(self.env_prefix, d, f.database, f.schema) for d in f.params.get('depends_on', [])],
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)

    def normalize_target_lag(self, target_lag: str):
        if target_lag.upper() == "DOWNSTREAM":
            return "DOWNSTREAM"

        num, _, unit  = target_lag.partition(" ")

        num = int(num)
        unit = unit.lower()

        if num > 1 and (unit in self.unit_singular_to_plural):
            unit = self.unit_singular_to_plural[unit]
        elif num == 1 and (unit in self.unit_plural_to_singular):
            unit = self.unit_plural_to_singular[unit]

        return f"{num} {unit}"
